/* autopano-win32, GUI
 * Copyright (C) 2004 -- Wei Zhen
 *
 * This program is free software released under the GNU General Public
 * License, which is included in this software package (doc/LICENSE).
 */

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;

namespace autopano_sift_winGUI
{
	/// <summary>
	/// FormProcess 的摘要说明。
	/// </summary>
	public class FormProcess : System.Windows.Forms.Form
	{
		private System.Windows.Forms.GroupBox groupBox1;
		private System.Windows.Forms.GroupBox groupBox2;
		public System.Windows.Forms.ProgressBar progressBarGeneration;
		public System.Windows.Forms.ProgressBar progressBarMatching;
		public System.Windows.Forms.Label labelGeneration;
		public System.Windows.Forms.Label labelMatching;
		public System.Windows.Forms.Button buttonOK;
		private System.Windows.Forms.GroupBox groupBox3;
		public System.Windows.Forms.ProgressBar refiningProgress;
		public System.Windows.Forms.Label refiningLabel;
		
		/// <summary>
		/// 必需的设计器变量。
		/// </summary>
		private System.ComponentModel.Container components = null;

		public FormProcess()
		{
			//
			// Windows 窗体设计器支持所必需的
			//
			InitializeComponent();

			//
			// TODO: 在 InitializeComponent 调用后添加任何构造函数代码
			//
		}

		/// <summary>
		/// 清理所有正在使用的资源。
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows 窗体设计器生成的代码
		/// <summary>
		/// 设计器支持所需的方法 - 不要使用代码编辑器修改
		/// 此方法的内容。
		/// </summary>
		private void InitializeComponent()
		{
			this.progressBarGeneration = new System.Windows.Forms.ProgressBar();
			this.progressBarMatching = new System.Windows.Forms.ProgressBar();
			this.groupBox1 = new System.Windows.Forms.GroupBox();
			this.labelGeneration = new System.Windows.Forms.Label();
			this.groupBox2 = new System.Windows.Forms.GroupBox();
			this.labelMatching = new System.Windows.Forms.Label();
			this.buttonOK = new System.Windows.Forms.Button();
			this.groupBox3 = new System.Windows.Forms.GroupBox();
			this.refiningProgress = new System.Windows.Forms.ProgressBar();
			this.refiningLabel = new System.Windows.Forms.Label();
			this.groupBox1.SuspendLayout();
			this.groupBox2.SuspendLayout();
			this.groupBox3.SuspendLayout();
			this.SuspendLayout();
			// 
			// progressBarGeneration
			// 
			this.progressBarGeneration.Location = new System.Drawing.Point(0, 15);
			this.progressBarGeneration.Name = "progressBarGeneration";
			this.progressBarGeneration.Size = new System.Drawing.Size(320, 21);
			this.progressBarGeneration.TabIndex = 0;
			// 
			// progressBarMatching
			// 
			this.progressBarMatching.Location = new System.Drawing.Point(0, 15);
			this.progressBarMatching.Name = "progressBarMatching";
			this.progressBarMatching.Size = new System.Drawing.Size(320, 21);
			this.progressBarMatching.TabIndex = 1;
			// 
			// groupBox1
			// 
			this.groupBox1.Controls.Add(this.labelGeneration);
			this.groupBox1.Controls.Add(this.progressBarGeneration);
			this.groupBox1.Location = new System.Drawing.Point(7, 7);
			this.groupBox1.Name = "groupBox1";
			this.groupBox1.Size = new System.Drawing.Size(320, 75);
			this.groupBox1.TabIndex = 2;
			this.groupBox1.TabStop = false;
			this.groupBox1.Text = "Keypoint generation";
			// 
			// labelGeneration
			// 
			this.labelGeneration.Location = new System.Drawing.Point(7, 45);
			this.labelGeneration.Name = "labelGeneration";
			this.labelGeneration.Size = new System.Drawing.Size(306, 21);
			this.labelGeneration.TabIndex = 1;
			this.labelGeneration.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// groupBox2
			// 
			this.groupBox2.Controls.Add(this.labelMatching);
			this.groupBox2.Controls.Add(this.progressBarMatching);
			this.groupBox2.Location = new System.Drawing.Point(8, 88);
			this.groupBox2.Name = "groupBox2";
			this.groupBox2.Size = new System.Drawing.Size(320, 74);
			this.groupBox2.TabIndex = 3;
			this.groupBox2.TabStop = false;
			this.groupBox2.Text = "Keypoint Matching";
			// 
			// labelMatching
			// 
			this.labelMatching.Location = new System.Drawing.Point(7, 45);
			this.labelMatching.Name = "labelMatching";
			this.labelMatching.Size = new System.Drawing.Size(306, 21);
			this.labelMatching.TabIndex = 2;
			this.labelMatching.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// buttonOK
			// 
			this.buttonOK.Location = new System.Drawing.Point(128, 240);
			this.buttonOK.Name = "buttonOK";
			this.buttonOK.Size = new System.Drawing.Size(63, 21);
			this.buttonOK.TabIndex = 4;
			this.buttonOK.Text = "OK";
			this.buttonOK.Click += new System.EventHandler(this.buttonExit_Click);
			// 
			// groupBox3
			// 
			this.groupBox3.Controls.Add(this.refiningProgress);
			this.groupBox3.Controls.Add(this.refiningLabel);
			this.groupBox3.Location = new System.Drawing.Point(8, 168);
			this.groupBox3.Name = "groupBox3";
			this.groupBox3.Size = new System.Drawing.Size(320, 72);
			this.groupBox3.TabIndex = 5;
			this.groupBox3.TabStop = false;
			this.groupBox3.Text = "Keypoint Refining";
			// 
			// refiningProgress
			// 
			this.refiningProgress.Location = new System.Drawing.Point(0, 16);
			this.refiningProgress.Name = "refiningProgress";
			this.refiningProgress.Size = new System.Drawing.Size(320, 21);
			this.refiningProgress.TabIndex = 3;
			// 
			// refiningLabel
			// 
			this.refiningLabel.Location = new System.Drawing.Point(0, 48);
			this.refiningLabel.Name = "refiningLabel";
			this.refiningLabel.Size = new System.Drawing.Size(320, 21);
			this.refiningLabel.TabIndex = 3;
			this.refiningLabel.TextAlign = System.Drawing.ContentAlignment.MiddleCenter;
			// 
			// FormProcess
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(333, 272);
			this.Controls.Add(this.groupBox3);
			this.Controls.Add(this.buttonOK);
			this.Controls.Add(this.groupBox2);
			this.Controls.Add(this.groupBox1);
			this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedDialog;
			this.Name = "FormProcess";
			this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
			this.Text = "PTO file generation";
			this.Load += new System.EventHandler(this.FormProcess_Load);
			this.groupBox1.ResumeLayout(false);
			this.groupBox2.ResumeLayout(false);
			this.groupBox3.ResumeLayout(false);
			this.ResumeLayout(false);

		}
		#endregion

		private void buttonExit_Click(object sender, System.EventArgs e)
		{
			Dispose();
		}

		private void FormProcess_Load(object sender, System.EventArgs e)
		{
		
		}

	}
}
