#!/bin/sh

rm -fR windows-bin

mkdir windows-bin
cp ../bin/*.exe windows-bin/
mv windows-bin/generatekeys-sd.exe windows-bin/generatekeys.exe
cp ../bin/*.dll windows-bin/
mkdir windows-bin/icons
cp ../util/autopano-win32/icons/*.png windows-bin/icons/

cp ../doc/LICENSE windows-bin/LICENSE.txt
cp ../CHANGES windows-bin/CHANGES.txt
cp ../doc/AUTOPANO-WINDOWS-README.txt windows-bin/WINDOWS-README.txt
flip -m windows-bin/*.txt
flip -m doc/*.txt

cp ../contrib/win-autopano-sift.vbs windows-bin/
cp ../contrib/win-autopano-sift-cmdline.vbs windows-bin/

wine "C:\\Program Files\\NSIS\\makensis.exe" installer.nsi

